#include"timer/timer.h"
#include"graph_interface.hpp"
#include"graph/graph.hpp"
#include"graph/boost_graphs.hpp"
#include"graph_interface.hpp"
#include"microarray.hpp"
#include"results.hpp"
#include"constants.h"
#include<interface.h>
#include<stdio.h>
#include<unistd.h>
#include<set>
#include<map>
#include<list>
#include<algorithm>
#include<string.h>
#include <boost/graph/adjacency_matrix.hpp>
#include <boost/graph/graph_utility.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/undirected_graph.hpp>
#include <boost/graph/bron_kerbosch_all_cliques.hpp>

namespace claim{

  using allinone::ArgAtoms;
  using allinone::ArgAtom;
  using allinone::Packages;
  using allinone::Genes;
  using allinone::Microarray;
  using allinone::Graph;
  using allinone::Results;

  struct clique_validator
  {
    deque<set<int> > * cliques;
    clique_validator(deque<set<int> > * cliques)
    { this->cliques = cliques; }
  
    template <typename Clique, typename Graph>
    void
    clique(const Clique& c, Graph& g)
    {
      cliques->push_back(set<int>());
      set<int> & s = cliques->back();
//      LOG(DEBUG, "Found clique: ");
      foreach(const typename Clique::value_type & v, c){
        int vertex = boost::get( boost::vertex_index_t(), g, v);
//        fprintf(stderr, "%d, ", vertex);
        s.insert(vertex); 
      }
//      fprintf(stderr, "\n");
//      cliques->back().insert(c.begin(), c.end());
    }
  };

  InterfaceFunctionBegin

    ArgAtoms outputs;
    deque<allinone::Multiple *> multiple;
    Interface * interface;
    int m = 51;

    for(ArgAtoms::iterator it = package->argNode.begin(); 
          it != package->argNode.end(); ++it){
      if((*it)->argAtomType & (ArgAtom::LongSwitch | ArgAtom::ShortSwitch)){
        if((*it)->value == "-i" || (*it)->value == "--input"){
          *++it;
          switch((*it)->argAtomType){
          case ArgAtom::ToProcess:
          case ArgAtom::PackageNumber:
            t.stop();
            allinone::getInterfaces(&multiple, *it, InterfaceFlow::Multiple);
            t.start();
            break;
          case ArgAtom::Arg:
            LOG(ERROR, "The package limit doesn't read files; Wrong argument: %s", 
                  (*it)->value.c_str());
          break;
          }
        }else if((*it)->value == "-o" || (*it)->value == "--output"){
          LOG(ERROR, "This package %s does not support outputs", 
                                                package->name.c_str());
        }else if((*it)->value == "-n" || (*it)->value == "--number-of"){
          n = atoi((*++it)->value.c_str());
          if((n != 1)){
            LOG(ERROR, "Wrong argument of \"--number-of\" switch: %d", 
                      (*it)->value.c_str());
          }
        }else if((*it)->value == "-m" || (*it)->value == "--min-multiplication"){
          m = atoi((*++it)->value.c_str());
        }
      }
    }

    if(multiple.size() != 2){
      LOG(ERROR, "Wrong number of input packages: %d,"
        " claim package, at this point of development process requires 2 sets"
        " of results", multiple.size());
    }

    if(multiple[0]->interfaces.size() < 1 || 
       multiple[1]->interfaces.size() < 1){
      LOG(ERROR, "claim requires that each set of results has at least 1"
      " result");
    }

    Results * refres = 0, * tmpres;
    deque<Results *> results_set[2];

    Results * ref = 0;
    for(int i = 0; i < 2; ++i){
      for(int j = 0, jend = multiple[i]->interfaces.size(); j < jend; ++j){
        if(multiple[i]->interfaces[j]->arg != InterfaceFlow::Results)
          LOG(ERROR, "Results are required in the claim package");
        tmpres = (Results *) multiple[i]->interfaces[j];
        results_set[i].push_back(tmpres);
        if(!refres) refres = tmpres;
        else{
          if(tmpres->get_no_vertices() != refres->get_no_vertices())
  LOG(ERROR, "There is inconsistency of the vertices within results sets");
          for(int i = 0, iend = tmpres->get_no_vertices(); i < iend; ++i){
            if(strcmp(tmpres->get_name(i), refres->get_name(i))){
  LOG(ERROR, "There is inconsistency of the vertices within results sets");
            }
          }
        }
      }
    }

    int no_verts = refres->get_no_vertices();

    graph::Graph<short>* g = new graph::Graph<short>(false, true, 1, no_verts);
    graph::Graph<char> * tmpg = new graph::Graph<char>(false, true, 1, no_verts); 

    g->set_names(&refres->get_name(0), no_verts);
    tmpg->set_names(&refres->get_name(0), no_verts);

    int no_subgraphs = 0;

    LOG(DEBUG, "Claim, parsed input parameters, allocated graphs");
  
    foreach(Results* res1, results_set[0])
    foreach(Results* res2, results_set[1]){
      if(res1->get_no_results() * res2->get_no_results() >= m){
        ++no_subgraphs;
        for(int i = 0; i < no_verts; ++i)
        for(int j = 0; j < no_verts; ++j)
          tmpg->set(i,j,0,0);  
        
        Results * ress[2] = {res1, res2};
        int andator = 1;
        typedef std::pair<Results**, Results**> resrespair;
        foreach(Results * res, resrespair(ress, ress + 2)){

          foreach(Results::Result & result, res->res){
            vector<int> v; v.reserve(result.second.size());
            typedef std::pair<const int,_PFloat> result_pair;
            foreach(result_pair & p, result.second)
              v.push_back(p.first);
            for(int i2 = 0; i2 < v.size(); ++i2)
            for(int j2 = i2+1; j2 < v.size(); ++j2){
              *tmpg->get_t(v[i2], v[j2], 0) |= andator;
              *tmpg->get_t(v[j2], v[i2], 0) |= andator;
            }
          }
          andator = andator << 1;
        }
        andator -= 1;
        for(int i2 = 0; i2 < no_verts; ++i2)
        for(int j2 = i2 + 1; j2 < no_verts; ++j2){
          if(!((~*tmpg->get_t(i2, j2, 0)) & andator)){
             *g->get_t(j2, i2, 0) = (*g->get_t(i2, j2, 0) += 1);
          }
        }
      }
    }

    LOG(DEBUG, "Claim have create the H graph, there where %d subgraphs",
                no_subgraphs);

    delete tmpg;
//    Graph * tmpg2 = graph::allocate_graph(false, true, 1, no_verts, "char", "adjacency_matrix");
//    tmpg2->set_names(&g->get_name(0), no_verts);

    Results * res = new Results(1, no_verts);
    res->set_names(&g->get_name(0), no_verts);

//    typedef boost::adjacency_matrix<boost::undirectedS > tmpgraph_t;
    typedef boost::undirected_graph<> tmpgraph_t;
    tmpgraph_t tmpadjm(no_verts) ;

    int level = 1;
    typedef deque<set<int> > cliques_t;
    vector<cliques_t> v(no_subgraphs);
    for(;no_subgraphs > 0; --no_subgraphs, ++level){
      for(int i2 = 0; i2 < no_verts; ++i2)
      for(int j2 = i2 + 1; j2 < no_verts; ++j2){
        if(*g->get_t(j2, i2, 0) == no_subgraphs){
          boost::add_edge(boost::vertex(i2, tmpadjm), 
                          boost::vertex(j2, tmpadjm), tmpadjm);
//          boost::add_edge(i2, j2, tmpadjm);
        }
      }
  boost::bron_kerbosch_all_cliques(tmpadjm, clique_validator(&v[level - 1]), 3);
      LOG(INFO, "Found %d cliques on level %d", v[level - 1].size(), level);
    }

    for(int i = 0; i < v.size() - 1; i++){
      foreach(set<int> & clique, v[i]){
        for(int j = i + 1; j < v.size(); j++){
          for(int k = v[j].size()-1; k >= 0; k--){
            set<int> & clique2 = v[j][k];
            if(clique.size() == clique2.size() && 
  std::includes(clique.begin(), clique.end(), clique2.begin(), clique2.end())){
              v[j].erase(v[j].begin() + k);
            }
          }
        }
        res->res.push_back(Results::Result(i + 1, map<int, _PFloat>()));
        Results::Result & result = res->res.back();
        foreach(int gene, clique)
          result.second.insert(std::pair<int,_PFloat>(gene, 0));
      }
      LOG(INFO, "Cliques %d on level %d after removal", v[i].size(), i + 1);
      
    }
    LOG(INFO, "Cliques %d on level %d after removal", v[v.size() - 1].size(), v.size());

    package->interface = res;
    package->interface->readOnly=true;
    package->interface->freeAble = false;

  InterfaceFunctionEnd

}
