#include"config.h"
#include"correlation.h"
#include"constants.h"
#include"graph/graph.hpp"
#include"microarray.hpp"
#include<stdio.h>
#include<deque>
#include<vector>
#include<set>
#include<list>
using namespace std;
#ifdef HAVE_TIMER_INTERFACE_H
#  include"timer/timer.h"  
#endif

using allinone::Microarray;


namespace corr {

  using allinone::ArgAtoms;
  using allinone::ArgAtom;
  using allinone::Packages;
  using allinone::Package;
  using allinone::Microarray;
  using allinone::Graph;
  

  void compute_pearson(Microarray *m, Graph *g, bool reverse_and_map02){
    bool allocated = false;
//    if(m->readOnly){
//      allocated = true;
//      m = new Microarray(m);
//    }
//    m->normalize();
    
    _PFloat sum;
    _PFloat * pptr1, * pptr2;
    int no_genes = m->get_no_genes();
    int vector_length = m->get_vector_length();
    _PFloat * avgs   = new _PFloat[no_genes];
    _PFloat * stdevs = new _PFloat[no_genes];

    for (int i = 0; i < no_genes; ++i){
      _PFloat sum = 0;
      for(int k = 0; k < vector_length; ++k)
        sum += m->get(i, k);
      avgs[i] = sum / vector_length;
      sum = 0;
      for(int k = 0; k < vector_length; ++k)
        sum += (m->get(i, k) - avgs[i]) * (m->get(i, k) - avgs[i]);
      stdevs[i] = sqrt(sum / vector_length);
    }

//    printf("Avgs:\n\t%g, %g, %g, %g, %g\n", avgs[0], avgs[1], avgs[2], avgs[3], avgs[4]);
//    printf("Stdevs:\n\t%g, %g, %g, %g, %g\n", stdevs[0], stdevs[1], stdevs[2], stdevs[3], stdevs[4]);

    for(int i = 0; i < no_genes; i++)
    for(int j = 0; j <= i; j++){
      sum = 0;
      pptr1 = &m->get(i,0);
      pptr2 = &m->get(j,0);
      for(int k = 0; k < vector_length; ++k)
        sum += (pptr1[k] - avgs[i]) * (pptr2[k] - avgs[j]);
      sum = sum / vector_length / stdevs[i] / stdevs[j];
      
//      if (i == j) printf("%d: %g\n", i, sum);

      if(reverse_and_map02){
        g->set(i,j,0, 1 - sum);
        g->set(j,i,0, 1 - sum);
      }else{
        g->set(i,j,0, sum);
        g->set(j,i,0, sum);
      }
    }
    if(allocated) delete m;
  }

  void compute_euclidean(Microarray *m, Graph *g){
    _PFloat sum, diff;
    _PFloat * pptr1, * pptr2;
    int no_genes = m->get_no_genes();
    int vector_length = m->get_vector_length();
    for(int i = 0; i < no_genes; i++)
    for(int j = 0; j < i; j++){
      sum = 0;
      pptr1 = &m->get(i,0);
      pptr2 = &m->get(j,0);
      for(int k = 0; k < vector_length; ++k){
        diff = pptr1[k] - pptr2[k];
        sum += diff * diff;
      }
      sum = sqrt(sum);
      g->set(i,j,0, sum);
      g->set(j,i,0, sum);
    }
  }

  InterfaceFunction{
    if(package->interface && package->interface->arg != InterfaceFlow::None)
      return 0;
    enum Correlation {pearson, euclidean} correlation = pearson;
    int n = 1;
    char str[100];
    sprintf(str, "%s:p[%d]", package->name.c_str(), package->packageNumber);
    timer::Timer t(str);
    t.start();
      deque<Microarray *> microarrays;
      Package * p;
      Interface * interface;
      bool reverse_and_map02 = false;

      for(ArgAtoms::iterator it = package->argNode.begin(); 
            it != package->argNode.end(); ++it){
        if((*it)->argAtomType & (ArgAtom::LongSwitch | ArgAtom::ShortSwitch)){
          if((*it)->value == "-i" || (*it)->value == "--input"){
            *++it;
            switch((*it)->argAtomType){
            case ArgAtom::ToProcess:
            case ArgAtom::PackageNumber:
              getInterfaces(&microarrays, (*it), InterfaceFlow::Microarray);
              break;
            case ArgAtom::Arg:
              LOG(ERROR, "Package corr does not read from files. Use microarray package instead.");
            break;
            }
          }else if((*it)->value == "-o" || (*it)->value == "--output"){
            LOG(ERROR, "Package corr does not write to files. Use appropriate graph package to write to the file");
          }else if((*it)->value == "-n" || (*it)->value == "--number-of"){
            n = atoi((*++it)->value.c_str());
            if(!(n >= 1)){
              LOG(ERROR, "Wrong argument of \"--number-of\" switch: %d", 
                        (*it)->value.c_str());
            }
          }else if((*it)->value == "-t" || (*it)->value == "--type"){
            if((*++it)->value == "euclidean")
              correlation = euclidean;
            else if((*++it)->value == "pearson")
              correlation = pearson;
            else LOG(ERROR, "Unsupported correlation type: %s", (*it)->value.c_str()); 
          } else if((*it)->value == "-r" || (*it)->value == "--revers-and-map-02"){
            reverse_and_map02 = true;
          }
        }
      }

      Microarray * microarray;
      list<Interface *> interfaces;
      switch(microarrays.size()){
      case 0:
        LOG(ERROR, "No input source specified for package[%d]: %s", 
          package->packageNumber, package->name.c_str());
      case 1:
        microarray = microarrays.front(); break;
      default:
        LOG(ERROR, "Too many input microarray sources specified for package[%d]: %s", 
          package->packageNumber, package->name.c_str());
      }

    Graph * graph = new graph::Graph<_PFloat>(false, true, 1, microarray->get_no_genes());
    graph->set_names(&microarray->get_name(0), microarray->get_no_genes(), 0);
    graph->set_descs(&microarray->get_desc(0), microarray->get_no_genes(), 0);

    graph->readOnly = true;
    graph->freeAble = false;
    
    switch(correlation){
      case euclidean:
        if(reverse_and_map02)
   LOG(ERROR, "The eucledian correlation cannot be used with reverse... switch");
        compute_euclidean(microarray, graph); break;
      case pearson:
        compute_pearson(microarray, graph, reverse_and_map02); break;
      default:
        LOG(ERROR, "Unsupported correlation type");
    }
    

    package->interface = graph;

    t.stop();
    return 0;
  }
}
