#include"timer/timer.h"
#include"graph_interface.hpp"
#include"genes.hpp"
#include"microarray.hpp"
#include"genes.hpp"
#include"graph/graph.hpp"
#include<stdio.h>
#include<unistd.h>
#include<set>
#include<map>
#include<list>
#include<algorithm>

namespace limit{

  using allinone::ArgAtoms;
  using allinone::ArgAtom;
  using allinone::Packages;
  using allinone::Package;
  using allinone::Genes;
  using allinone::Microarray;
  using allinone::Graph;
 
  void visit(Graph * g, std::set<int> * verts, int vert, std::vector<int> * names_to_i = 0){
    verts->insert(vert);
    for(int i = 0, iend = g->get_no_vertices(); i < iend; i++){
      if((!names_to_i || names_to_i->at(i)) && g->is_edge(vert, i, 0)) {
        std::pair<std::set<int>::iterator,bool> p = verts->insert(i);
        if(p.second) visit(g, verts, i, names_to_i);
      }
    }
  }

  struct sets_less{ 
    int reverse;
    sets_less(int reverse) : reverse (reverse){ }
    bool operator() (const std::set<int> * _1, const std::set<int> * _2) const{
      if(reverse) return _1->size() > _2->size();
      return _1->size() < _2->size();
    }
  };

  void span(Graph * g, std::set<const char *, allinone::charptr_less> & names,
      std::set<const char *, allinone::charptr_less> & largest_subgraph_names){
    std::vector<int> names_to_i(g->get_no_vertices(), 0);
    std::vector<int> visited(g->get_no_vertices(), 0);

    //span subgraphs
    std::deque<std::set<int>*> sets;
    std::vector<int> itoi(g->get_no_vertices(), -1);
    
    if(names.size()){
      foreach(const char * name, names)
        names_to_i[g->get_no(name)] = 1;
    }else foreach(int & i, names_to_i)
            i = 1;
    
    for(int i = 0, iend = g->get_no_vertices(); i < iend; i++){
      if(names_to_i[i] && ! visited[i]){
        std::set<int>  * spanned = new std::set<int>();
        visit(g, spanned, i, &names_to_i);
        if(spanned->size() == 1) delete spanned;
        else{
          std::set<int> existing;
          foreach(int j, *spanned){
            if(itoi[j] != -1) existing.insert(j);
            visited[j] = true;
          }
          int set_i = sets.size();
          if(existing.size() > 1) 
            LOG(ERROR, "Not implemented finding span subgraph in directed graphs, sorry...");
          else if(existing.size() == 1) {
            set_i = *existing.begin();
            sets[set_i]->insert(spanned->begin(), spanned->end());
            delete spanned;
          }
          else {
            set_i = sets.size();
            sets.push_back(spanned);
          }
          foreach(int j, *spanned)
            itoi[j] = set_i;
        }
      }
    }
  
    if(sets.size()){
      sort(sets.begin(), sets.end(), sets_less(1));
      foreach(int i, *sets.front())
        largest_subgraph_names.insert(g->get_name(i));
    }
    foreach(std::set<int> * s, sets)
      delete s;

  }

  InterfaceFunction{

    if(package->interface && package->interface->arg != InterfaceFlow::None)
      return 0;
    int n = 1;
    char str[100];
    sprintf(str, "%s:p[%d]", package->name.c_str(), package->packageNumber);
    timer::Timer t(str);
    t.start();
      ArgAtoms outputs;
      deque<allinone::Graph *> graphs;
      deque<allinone::Microarray *> mas;
      deque<Interface *> ins;
      Package * p;
      Interface * interface;

      bool span = false;
      char delimit = ';';
      deque<allinone::Genes *> gns;

      for(ArgAtoms::iterator it = package->argNode.begin(); 
            it != package->argNode.end(); ++it){
        if((*it)->argAtomType & (ArgAtom::LongSwitch | ArgAtom::ShortSwitch)){
          if((*it)->value == "-i" || (*it)->value == "--input"){
            *++it;
            switch((*it)->argAtomType){
            case ArgAtom::ToProcess:
            case ArgAtom::PackageNumber:
              getInterfaces(&ins, (*it), 
InterfaceFlow::InterfaceArg(InterfaceFlow::Graph | InterfaceFlow::Microarray));
            break;

            case ArgAtom::Arg:
              LOG(WARNING, "The package limit doesn't read files; Wrong argument: %s", 
                    p->name.c_str());
            }
          }else if((*it)->value == "-o" || (*it)->value == "--output"){
            outputs.push_back(*++it);
          }else if((*it)->value == "-s" || (*it)->value == "--span"){
            span=true;
          }else if((*it)->value == "-t" || (*it)->value == "--to"){
            *++it;
            switch((*it)->argAtomType){
            case ArgAtom::ToProcess:
            case ArgAtom::PackageNumber:
              getInterfaces(&gns, (*it), 
InterfaceFlow::InterfaceArg(InterfaceFlow::Graph | InterfaceFlow::Microarray));
            break;
            case ArgAtom::Arg:
              LOG(WARNING, "The package limit doesn't read files; Wrong argument: %s", 
                    (*it)->value.c_str());
            }
          }else if((*it)->value == "-n" || (*it)->value == "--number-of"){
            n = atoi((*++it)->value.c_str());
            if(!(n >= 1)){
              LOG(ERROR, "Wrong argument of \"--number-of\" switch: %d", 
                        (*it)->value.c_str());
            }
          }
        }
      }
      
      foreach(Interface * i, ins){
        if(i->arg & InterfaceFlow::Graph) graphs.push_back((Graph *)i);
        else if(i->arg & InterfaceFlow::Microarray) 
          mas.push_back((Microarray *)i);
      }

      if(graphs.size() + mas.size() != 1)
        LOG(ERROR, "Wrong number of input packages, 1 is required: ", 
          graphs.size() + mas.size());

      Genes * old = 0;
      if(mas.size())          old = (Genes *) mas.front();
      else if(graphs.size())  old = (Genes *) graphs.front();

      std::set<const char *, allinone::charptr_less> names;
      std::set<const char *, allinone::charptr_less> common_names;
      std::set<const char *, allinone::charptr_less> oldnames;
      if(gns.size()){
        foreach(allinone::Genes *g, gns){
          int n_verts = g->get_no_vertices();
          const char ** name = &g->get_name(0);
          names.insert(name, name + n_verts);
        }
      }

      if(names.size()){
        const char ** tmppp = &old->get_name(0);
        oldnames.insert(tmppp, tmppp + old->get_no_vertices());
        std::set_intersection(names.begin(), names.end(), oldnames.begin(), 
          oldnames.end(), std::inserter(common_names, common_names.begin()), 
          allinone::charptr_less());
        if(!common_names.size())
          LOG(WARNING, "The intersection of genes names is empty; Leaving graph unchanged");
      }

      if(span){
        if(mas.size()) LOG(WARNING, "Can't compute span graph for MA");
        else{
          std::set<const char *, allinone::charptr_less> common_names2;
          if(common_names.size()){
            common_names2.insert(common_names.begin(), common_names.end());
            common_names.clear();
          }
          limit::span(graphs.front(), common_names2, common_names);
          if(!common_names.size()) 
            LOG(WARNING, "The span graph size is equal to 0; Leaving graph unchanged;");
        }
      }

      if(!common_names.size()){
        LOG(WARNING, "The graph/MA is unchaged after limiting the names");
        if(mas.size()) package->interface = mas.front();
        else package->interface = graphs.front();
      } else {
        if(graphs.size()){
          Graph * g = graphs.front();
          interface = g = graph::allocate_graph(g->get_is_directed(), 
                  g->get_is_weighted(), g->get_vector_length(), 
                  common_names.size(), g->get_datatype_s(), 
                  g->get_graphtype_s(), g->get_threshold());
          old = g;
        }else {
          Microarray * m = mas.front();
          interface = m = new 
                Microarray(m->get_vector_length(), common_names.size());
          old = m;
        }
        int i = 0;
        foreach(const char * str, common_names)
          old->set_names(&str, 1, i++);

        if(graphs.size()){
          Graph * g = (Graph *) interface;
          Graph *og = graphs.front();
          for(int i = 0, iend = g->get_no_vertices(); i != iend; i++){
            int from = og->get_no(g->get_name(i)); 
            for(int j = 0; j != iend; j++){
              int to = og->get_no(g->get_name(j)); 
              for(int k = 0, kend = g->get_vector_length(); k != kend; k++){
                double tmp = og->get(from, to, k);
                if(tmp != 0) g->set(i, j, k, tmp);
              }
            }
          }
        } else {
          Microarray * m = (Microarray*) interface;
          Microarray *om = mas.front();
          for(int i = 0, iend = m->get_no_vertices(); i != iend; i++){
            int from = om->get_no(m->get_name(i)); 
            for(int k = 0, kend = m->get_vector_length(); k != kend; k++){
              double tmp = om->get(from, k);
              if(tmp != 0) m->get(i, k) = tmp;
            }
          }
          m->set_exps(&om->get_exp(0), om->get_vector_length());
        }
        package->interface = interface;
      } 
      
      package->interface->readOnly=true;
      package->interface->freeAble = false;

    t.stop();
    return 0;
  }
}
