#include "logger.h"
#include "interface.h"
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#if HAVE_EXECINFO_H
# include"stacktrace.h"
#endif

namespace logger {
  
  #define FILENAME_WIDTH 25

  static const char * verbosity2str[] = {
    "", "ERR", "WAR", "NOT", "INF", "DEB"
  };


  int log(Verbosity error, const char * filename, const int line, const char * str, ...){
    if(error <= verbosity) {
      static char * output  = (char *)malloc(10000);
      static char * output2 = (char *)malloc(10000);
      static int stringsize = 10000;

      va_list(argp);
      va_start(argp, str);
      
      int outlen = vsnprintf(output,stringsize,str,argp);
      if(outlen >= stringsize - 100){
        stringsize = 2*outlen; 
        output  = (char *)malloc(10000);
        output2 = (char *)malloc(10000);
        vsnprintf(output,stringsize,str,argp);
      }

      int infolen = FILENAME_WIDTH + 5 + 5;
      int filelen = strlen(filename);
      int i = 0, j = 0;

      if(filelen > FILENAME_WIDTH)
        j+=sprintf(&output2[j], "[%s]...%s:%04d:", verbosity2str[error], 
          &filename[filelen - FILENAME_WIDTH + 3], line);
      else
        j+=sprintf(&output2[j], "[%s]%" STRINGIFY(FILENAME_WIDTH) "s:%04d:", 
                                  verbosity2str[error], filename, line);

      for(i = 0; output[i]; i++, j++){
        output2[j] = output[i];
        switch (output[i]){
          case '\n':{
            if(!output[i + 1]) continue;
            for(int k=0; k < infolen; k++)
              output2[++j]=' ';
            output2[++j]='|';
          }break;
        }
      }
      output2[j] = '\0';
      if(output2[j - 1] != '\n') fprintf(stderr, "%s\n", output2);
      else fprintf(stderr, "%s", output2);
      if(error == ERROR){
#if HAVE_EXECINFO_H
        print_stacktrace();
#endif
        exit (-1);
      }
        
    }
    return 0;
  }


}
