#include"config.h"
#include"microarray.hpp"
#include"constants.h"
#include<stdio.h>
#include<deque>
#include<vector>
#include<set>
#include<list>
using namespace std;
#ifdef HAVE_TIMER_INTERFACE_H
#  include"timer/timer.h"  
#endif

using allinone::Microarray;

namespace microarray {


  void test(){
    Microarray m(4, 10);

    const char * names[] = {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"};
    m.set_names(names, 10);
    
    LOG(INFO, "Gene names:", "");

    int j = 0;
    for(Microarray::iterator i = m.begin(), 
                             iend = m.end();
                                i!=iend; ++i){
      fprintf(stderr, "%s, ", i->get_name());
      (*i)[0] = j+=2;
    }
    fprintf(stderr, "\n");
    LOG(NOTICE, "1: %f, 9:%f\n\ttest", m[1][0], m[9][0]);

    LOG(ERROR, "BLAD", "");
  }

  using allinone::ArgAtoms;
  using allinone::ArgAtom;
  using allinone::Packages;
  using allinone::Package;
  using allinone::Microarray;
  
  void save_file(string name, Microarray * m, char delimit){
    if(!m) return;
    string name2 = allinone::getWriteFilename(name);
    FILE * f = name2 == "-" ? stdout : fopen(name2.c_str(), "w");
    if(!f){
      LOG(WARNING, "Can't open file for write: %s, skipping", name2.c_str());
      return;
    }
    int columns = m->get_vector_length();
    int genes = m->get_no_genes();
    fprintf(f, "Gene%cDescription", delimit);
    for(int i = 0; i < columns; ++i)
      fprintf(f, "%c%s", delimit, m->get_exp(i));

    for(int i = 0; i < genes; ++i){
      fprintf(f, "\n%s%c%s", m->get_name(i)?:"", delimit, m->get_desc(i)?:"");
      for(int j = 0; j < columns; ++j){
        fprintf(f, "%c%.20g", delimit, m->get(i, j));
      }
    }
    if(name2 != "-") fclose(f);
  }

  Microarray * read_file(string name, char delimit){
    deque<string> * deq = allinone::getStringsFromFile(name.c_str());
    if(!deq) LOG(ERROR, "Can't read file: %s", name.c_str());
    size_t lines = deq->size() ;
    if(lines < 2) LOG(ERROR, "File: %s has less than two lines", name.c_str());
    
    const char * tmpstr; 
    size_t no = -1;   
    size_t no_floats = 0;   
    list<string> * l;
    l = allinone::splitNStrip((*deq)[0], delimit);
    no = l->size();
    if(no <= 2){
      LOG(ERROR, "First decoded line: %s, has only one element.\n" 
          "At least 3 are required: name + 2 experimental results;", 
          (*deq)[0].c_str());
    }
    delete l;
    
    l = allinone::splitNStrip((*deq)[1], delimit);
    if(no != l->size())
      LOG(ERROR, "Second decoded line: %s, has different number of elements (%ld) than the first one (%ld).\n",
          (*deq)[1].c_str(), l->size(), no);
    rforeach(string & s, *l){ 
      if(allinone::toFloat(s)) no_floats++;
      else break;
    }
    if(!(no - no_floats >= 1 && no - no_floats <=2))
      LOG(ERROR, "There are %ld column descriptions. 1 or 2 are allowed.", 
            no - no_floats);
    if(no_floats < 2) LOG(ERROR, "At least 2 floats are required");
    LOG(INFO, "File %s, has %ld columns and %ld of them are experiments", 
            name.c_str(), no, no_floats);
    delete l;

    rforeach(string & str, *deq){
      l = allinone::splitNStrip((*deq)[1], delimit);
      int no_floats2 = 0;
      rforeach(string & str, *l){
        if(allinone::toFloat(str)) no_floats2++;
        else break;
      }
      if(no_floats2 == no_floats) break;
      else if (no_floats2) 
          LOG(NOTICE, "Skipping line: %s, wrong number of floats", str.c_str());
      lines --;
      delete l;
    }

    lines--;
    if(lines < 2) LOG(ERROR, "Less than 2 lines in the file: %s", name.c_str());

    Microarray *m = new Microarray(no_floats, lines);
    int i = -1;

    foreach(string & str, *deq){
#ifdef __DEBUG__
//      LOG(DEBUG, "Read Line: %s", str.c_str());
#endif /*__DEBUG__*/
      list<string> * l = allinone::splitNStrip(str, delimit);
      int no_floats2 = 0;
      double d;
      if(i >= 0){
        list<string>::iterator it = l->begin();
        m->set_names(&(tmpstr = it++->c_str()), 1, i);
        if(no - no_floats == 2) m->set_descs(&(tmpstr = it++->c_str()), 1, i);
        for( ; it!=l->end() && no_floats2 < no_floats; ++it){ 
          if(!allinone::toFloat(*it, &(*m)[i][no_floats2++])){
            LOG(WARNING, "Could not convert to float: %s in file: %s", 
                      it->c_str(), name.c_str());
          }
//          LOG(DEBUG, "Converted %s, to: %f", it->c_str(), (*m)[i][no_floats2 - 1]);
        }
        if(no_floats2 != no_floats) 
          LOG(WARNING, "file: %s, line: %s, wrong number of floats: %ld, expected: %ld",
              name.c_str(), str.c_str(), no_floats2, no_floats);
        if(it != l->end()) 
          LOG(WARNING, "file: %s, line: %s, wrong number of columns: %ld, expected: %ld",
            l->size(), no);
      }else{
        vector<const char *> v(no);
        int j = 0;
        foreach (string & col, *l) v[j++] = col.c_str();
        m->set_exps(&v[no - no_floats], no_floats, 0);
      }
      i++;
      delete l;
    }

    delete deq;
    return m;
  }


  Microarray * merge(deque<Microarray*> & l){
    set<const char *, allinone::charptr_less> genes;
    vector<const char *> exps; exps.reserve(20);
    foreach(Microarray * m, l){
      const char ** strpp = &m->get_name(0);
      genes.insert(strpp, &strpp[m->get_no_genes()]);
      strpp = &m->get_exp(0);
      exps.insert(exps.end(), strpp, &strpp[m->get_vector_length()]);
    }
    if(!genes.size() || !exps.size()) return 0;

    Microarray * mnew = new Microarray(exps.size(), genes.size());
    mnew->set_exps(&exps[0], exps.size());
    mnew->set_names(&vector<const char *>(genes.begin(), genes.end())[0], genes.size());

    foreach(Microarray * m, l){
      for(int i = 0, iend = m->get_no_genes(); i < iend; ++i){
        const char * name = m->get_name(i);
        for(int j = 0, jend = m->get_vector_length(); j < jend; ++j){
          mnew->get(name, m->get_exp(j)) = m->get(i, j);
        }
      }
    }

    return mnew;
  }


  InterfaceFunction{
    if(package->interface && package->interface->arg != InterfaceFlow::None)
      return 0;
    int n = 1;
    char str[100];
    sprintf(str, "%s:p[%d]", package->name.c_str(), package->packageNumber);
    timer::Timer t(str);
    t.start();
      ArgAtoms outputs;
      deque<Microarray *> microarrays;
      Interface * interface;
      char delimit = ';';

      for(ArgAtoms::iterator it = package->argNode.begin(); 
            it != package->argNode.end(); ++it){
        if((*it)->argAtomType & (ArgAtom::LongSwitch | ArgAtom::ShortSwitch)){
          if((*it)->value == "-i" || (*it)->value == "--input"){
            *++it;
            switch((*it)->argAtomType){
            case ArgAtom::ToProcess:
            case ArgAtom::PackageNumber:
              getInterfaces(&microarrays, (*it), InterfaceFlow::Microarray);
              break;
            case ArgAtom::Arg:
              microarrays.push_back(read_file((*it)->value, delimit));
            break;
            }
          }else if((*it)->value == "-o" || (*it)->value == "--output"){
            outputs.push_back(*++it);
          }else if((*it)->value == "-n" || (*it)->value == "--number-of"){
            n = atoi((*++it)->value.c_str());
            if(!(n >= 1)){
              LOG(ERROR, "Wrong argument of \"--number-of\" switch: %d", 
                        (*it)->value.c_str());
            }
          }else if((*it)->value == "-d"){
            delimit = allinone::parseDelimiter((*++it)->value);
          }
        }
      }

      Microarray * microarray;
      list<Interface *> interfaces;
      switch(microarrays.size()){
      case 0:
        LOG(ERROR, "No input source specified for package[%d]: %s", 
          package->packageNumber, package->name.c_str());
      case 1:
        microarray = microarrays.front(); break;
      default:
        microarray = merge(microarrays);
        foreach(Microarray * m, microarrays) 
          interfaces.push_back((Interface *) m);
        allinone::freeFreeable(interfaces);
      }

    microarray->readOnly = true;
    microarray->freeAble = false;

    package->interface = microarray;

    foreach(ArgAtom * o, outputs)
      save_file(o->value, microarray, delimit);
    
    t.stop();
    return 0;
  }
}
