#include"config.h"
#include"graph/graph.hpp"
#include"graph/boost_graphs.hpp"
//#include"graph_interface.hpp"
#include"constants.h"
#include<stdio.h>
#include<deque>
#include<vector>
#include<set>
#include<list>
#include<math.h>
#include<boost/preprocessor.hpp>
#include<boost/multi_array.hpp>
#include<boost/graph/adjacency_list.hpp>
#include<boost/graph/adjacency_matrix.hpp>
#include<boost/graph/johnson_all_pairs_shortest.hpp>
using namespace std;
#ifdef HAVE_TIMER_INTERFACE_H
#  include"timer/timer.h"  
#endif

namespace shortest_path {

  
  using allinone::ArgAtoms;
  using allinone::ArgAtom;
  using allinone::Packages;
  using allinone::Package;
  using allinone::Graph;
  using timer::Timer;
  using boost::vecS;
  using boost::mapS;
  using boost::adjacency_list;
  using boost::adjacency_matrix;
  using boost::directedS;
  using boost::undirectedS;
  using boost::no_property;

  template<class T>
  void trim(allinone::Graph * g, T lower = 0, T upper = 13, bool set_upper = false){
    long no_verts = g->get_no_vertices();    
    if(set_upper){
      for(long j = 0; j < no_verts; j++)
      for(long i = 0; i < no_verts; i++){
        T val = g->get(j, i, 0);
        if(val > upper || val < lower)
          g->set(j, i, 0, (double)upper);
      }
    }
    for(long i = 0; i < no_verts; i++){
      g->set(i, i, 0, 0);
    }
  }

  template<class BGraph, class T>
  void shortest_path_boost(BGraph * g, graph::Graph<T> * gout){
    int tmp[] = {gout->get_no_vertices(), gout->get_no_vertices()};
    boost::multi_array_ref<T, 2> D(gout->get_t(0,0,0), std::list<int>(tmp, tmp+2));
    try
    {
      bool x = boost::johnson_all_pairs_shortest_paths(*g, D 
          ,boost::weight_map(boost::get(&graph::Edge_property<T, 1>::p, *g)));
      if(!x) LOG(WARNING, "Johnson shortest path algorythm reported failure...");
    }
    catch (std::bad_alloc& ba)
    {
      LOG(ERROR, "Caught bad_alloc: %s", ba.what());
    }
  }

  struct HeapCmp {
    Graph * g;
    int from;
    bool operator() (const int &a, const int &b)
    {
      return (g->get(from, a, 0) < g->get(from, b, 0)) ? true : 
             (g->get(from, a, 0) > g->get(from, b, 0)) ? false: a<b;
    }
    HeapCmp(Graph * g, int from) : g(g), from(from) { }
    HeapCmp() { }
  };

  typedef set<int, HeapCmp> Heap;
  
  template<class T>
  void shortest_path(Graph * gin, graph::Graph<T> * gout){
    size_t genes = gin->get_no_vertices();
    size_t elements = genes * genes;

    Timer t("setting data");
    t.start();
    LOG(INFO, "Computing shortest path");
    T * pptr;
    for(size_t i = 0; i < genes; i++){ 
      for(size_t j = 0; j < genes; j++){ 
        if(i != j){
          pptr = gout->get_t(i,j,0);
          if(*pptr == 0) *pptr++ = std::numeric_limits<T>::max();
        }
      }
    }
    t.stop();

    Timer t2("Heap alloc");
    t2.start();
    vector<Heap *> heaps(gout->get_no_vertices());
    for (int i = 0; i < genes; i++){ 
//      Vertex<_PFloat> heapv(gout, i);
      set<int, HeapCmp> * heap = new set<int, HeapCmp>(HeapCmp(gout, i));        
      for (int j = 0; j < genes; j++) {
        if(i != j && !(std::numeric_limits<T>::max() == gout->get(i, j, 0))) heap->insert(j);
      }
      heaps[i] = heap;
    }
    t2.stop();

    for(int i = 0; i < genes; i++){
      if(!(i%100))
        LOG(INFO, "Processing gene: %d", i);
//      Vertex<_PFloat> heapv(gout, i);
      Heap * heap = new Heap(heaps[i]->begin(), heaps[i]->end(), HeapCmp(gout, i));
      Heap * heap2;        
      Heap::iterator heapit;        
       
      int u;
      _PFloat w;
      
      Timer t4("Heap operat");
      Timer t5("Gene loop");
      Timer t6("other");
      t5.start();
      t4.start();
      while(!heap->empty())
      {
        u = *(heapit = heap->begin());
        heap->erase(heapit);
        heap2 = heaps[u];
        t4.stop();
      
        t6.start();
        foreach(int j, *heap2)
//        for(int j = i + 1; j < genes; j++)
        {
          w = gout->get(u, j, 0);
          if(w == std::numeric_limits<T>::max()) continue;
          if (gout->get(i, u, 0) + w < gout->get(i, j, 0))
          {
            t6.stop();
            
            t4.start();
            heapit = heap->find(j);
            if(heapit != heap->end()) heap->erase(heapit);
            t4.stop();
            t6.start();
            gout->set(i, j, 0, gout->get(i, u, 0) + w);
//            heapv[j][0] = heapv[u][0] + w;
            t6.stop();
            t4.start();
            heap->insert(j);
            t4.stop();
            t6.start();
          }
        }
        t6.stop();
        t4.start();
      }
//      for(int j = i + 1; j < genes; j++)
//        heap->insert(j);
      t5.stop();

      t4.stop();
      Timer t3("Heap dealloc");
      t3.start();
      delete heap;
      t3.stop();
    }
    
    Timer t7("scan");
    t7.start();
    for (int i = 0; i < genes; i++) 
    for (int j = 0; j < genes; j++) {
      pptr = gout->get_t(i, j, 0);
      if(*pptr == std::numeric_limits<T>::max()) *pptr = 0;
    }
    t7.stop();

    foreach(Heap * h, heaps) delete h;
//    sleep(300);
  }
  
  InterfaceFunction{
    if(package->interface && package->interface->arg != InterfaceFlow::None)
      return 0;
    int n = 1;
    char str[100];
    sprintf(str, "%s:p[%d]", package->name.c_str(), package->packageNumber);
    timer::Timer t(str);
    t.start();
      ArgAtoms outputs;
      deque<Graph *> graphs;
      Package * p;
      Interface * interface;
      double lower = 0;
      double upper = 13;
      bool set_upper = false;

      for(ArgAtoms::iterator it = package->argNode.begin(); 
            it != package->argNode.end(); ++it){
        if((*it)->argAtomType & (ArgAtom::LongSwitch | ArgAtom::ShortSwitch)){
          if((*it)->value == "-i" || (*it)->value == "--input"){
            *++it;
            switch((*it)->argAtomType){
            case ArgAtom::ToProcess:
            case ArgAtom::PackageNumber:
              getInterfaces(&graphs, (*it), InterfaceFlow::Graph);
              break;
            
            case ArgAtom::Arg:
              LOG(WARNING, "The shortest path package can only take graphs from other packages");
//              graphs.push_back(read_file((*it)->value));
            break;
            }
          }else if((*it)->value == "-o" || (*it)->value == "--output"){
            LOG(WARNING, "The shortest path package cannot output to file anything;");
//            outputs.push_back(*++it);
            ++it;
          }else if((*it)->value == "-n" || (*it)->value == "--number-of"){
            n = atoi((*++it)->value.c_str());
            if(!(n >= 1)){
              LOG(ERROR, "Wrong argument of \"--number-of\" switch: %d", 
                        (*it)->value.c_str());
            }
          }else if((*it)->value == "-u" || (*it)->value == "--upper-thres"){
            if(allinone::toFloat((*++it)->value, &upper))
              LOG(ERROR, "Cannot convert %s to float", (*it)->value.c_str());
          }else if((*it)->value == "-l" || (*it)->value == "--lower-thres"){
            if(allinone::toFloat((*++it)->value, &lower))
              LOG(ERROR, "Cannot convert %s to float", (*it)->value.c_str());
          }else if((*it)->value == "-s" || (*it)->value == "--set-upper"){
            set_upper = true;
          }
        }
      }

      Graph * gr;
      list<Interface *> interfaces;
      switch(graphs.size()){
      case 0:
        LOG(ERROR, "No input source specified for package[%d]: %s", 
          package->packageNumber, package->name.c_str());
      case 1:
        gr = graphs.front(); break;
      default:
        LOG(ERROR, "Currently PPI package supports only one input graph.\n"
                   "You can merge them in Graph package.");
      }

      gr->readOnly = true;
      gr->freeAble = false;

      Graph * gnew = 0;

      switch(gr->get_graphtype()){ 
        case Graph::Graph_matrix: 
#define DATATYPE_IF(r, x, t)                                                  \
        if(gr->get_datatype() == Graph::DATATYPES_HELPER2(datatype_, t)){     \
          graph::Graph<t> * gnews =                                           \
                   new graph::Graph<t>(gr, gr->get_no_vertices());            \
          shortest_path<t>(gr,  gnews);                                       \
          gnew = gnews;                                                       \
        } else 
        BOOST_PP_SEQ_FOR_EACH(DATATYPE_IF, ~, DATATYPES) 
          LOG(ERROR, "Undefined data-type: %s", gr->get_datatype_s().c_str());
#undef DATATYPE_IF
         break;
#define DATATYPE_IF(r, x, t)                                                  \
        if(gr->get_datatype() == Graph::DATATYPES_HELPER2(datatype_, t)){     \
          graph::Graph<t> * gnews =                                           \
            new graph::Graph<t>(gr, gr->get_no_vertices(), false);            \
          if(gr->get_is_directed())                                           \
shortest_path_boost((BASE directedS,no_property, graph::Edge_property<t, 1> >*)gr->graph(),  gnews);\
          else                                                                \
shortest_path_boost((BASE undirectedS,no_property, graph::Edge_property<t, 1> >*)gr->graph(),  gnews);\
          gnew = gnews;                                                       \
        } else 
        case Graph::Graph_adjacency_list: 
#define BASE adjacency_list<mapS,vecS,
        BOOST_PP_SEQ_FOR_EACH(DATATYPE_IF, ~, DATATYPES)
          LOG(ERROR, "Undefined data-type: %s", gr->get_datatype_s().c_str());
#undef BASE 
        break;
        case Graph::Graph_adjacency_matrix: 
#define BASE adjacency_matrix<
        BOOST_PP_SEQ_FOR_EACH(DATATYPE_IF, ~, DATATYPES) 
          LOG(ERROR, "Undefined data-type: %s", gr->get_datatype_s().c_str());
#undef BASE
        break;
#undef DATATYPE_IF
      }

#define DATATYPE_IF(r, x, t)                                                  \
      if(gr->get_datatype() == Graph::DATATYPES_HELPER2(datatype_, t)){       \
        trim<t>(gnew, lower, upper, set_upper);                               \
      } else 
      BOOST_PP_SEQ_FOR_EACH(DATATYPE_IF, ~, DATATYPES) 
        LOG(ERROR, "Undefined data-type: %s", gr->get_datatype_s().c_str());
#undef DATATYPE_IF

      package->interface = gnew;
    t.stop();
    return 0;
  }
}
